defmodule ClickHouse.Query do
  @moduledoc """
  A ClickHouse query.
  """

  alias ClickHouse.{DataType, Query.Statement}

  defstruct [
    :format,
    :statement
  ]

  @typedoc """
  A ClickHouse query
  """
  @type t :: %__MODULE__{
          format: ClickHouse.Format.t() | nil,
          statement: iolist()
        }

  ################################
  # Public API
  ################################

  @doc """
  Prepares a new ClickHouse query.
  """
  @spec prepare(ClickHouse.Client.t(), ClickHouse.statement(), ClickHouse.params()) ::
          ClickHouse.Query.t()
  def prepare(client, statement, params \\ [])

  def prepare(client, statement, params) when is_binary(statement) do
    statement = Statement.new(statement)
    prepare(client, statement, params)
  end

  def prepare(client, %Statement{} = statement, params) do
    format = Map.get(client.formats, statement.format)
    statement = maybe_encode(format, statement, params)
    %__MODULE__{format: format, statement: statement}
  end

  ################################
  # Private API
  ################################

  defp maybe_encode(_, statement, []) do
    statement.tokens
  end

  defp maybe_encode(nil, %Statement{command: :insert}, [_]) do
    raise ArgumentError, """
    Invalid insert statement - no encoding format available. \
    Please see ClickHouse.Format documentation for more information.
    """
  end

  defp maybe_encode(_format, %Statement{command: :insert}, [_, _ | _]) do
    raise ArgumentError, """
    Insert query cannot contain more than a single parameter representing all rows.
    """
  end

  defp maybe_encode(format, %Statement{command: :insert, tokens: [statement, _]}, [rows]) do
    params = format.encode(rows)
    [statement, params]
  end

  defp maybe_encode(_, %Statement{command: command} = statement, params)
       when command in [:select, :create, :alter, :unknown] do
    encode(statement.tokens, params)
  end

  defp encode(parts, params, acc \\ [])

  defp encode([], _, acc), do: acc

  defp encode(["?" | _], [], _acc) do
    raise ArgumentError, """
    Query parameters do not match the number of embedded '?'.
    """
  end

  defp encode(["?"], [data], acc) do
    [acc | [DataType.encode(data)]]
  end

  defp encode(["?" | parts], [data | params], acc) do
    encode(parts, params, [acc | [DataType.encode(data)]])
  end

  defp encode([part], _params, acc) do
    [acc | [part]]
  end

  defp encode([part | parts], params, acc) do
    encode(parts, params, [acc | [part]])
  end

  defimpl String.Chars do
    def to_string(%{statement: statement}) do
      IO.iodata_to_binary(statement)
    end
  end

  defimpl Inspect do
    def inspect(%{statement: statement}, _) do
      "#ClickHouse.Query<\"#{IO.iodata_to_binary(statement)}\">"
    end
  end
end
