defmodule ClickHouse.Query.Statement do
  @moduledoc false

  @values_regex ~r/\bVALUES|values\b/i
  @format_regex ~r/(?<= (?:FORMAT|format) )(\w+)/

  defstruct [
    :tokens,
    :command,
    :format
  ]

  @type command :: :create | :select | :insert | :alter | :unknown

  @type t :: %__MODULE__{
          tokens: iodata(),
          command: command(),
          format: ClickHouse.Format.name()
        }

  @spec new(binary()) :: ClickHouse.Query.Statement.t()
  def new(statement) do
    statement = strip_newlines(statement)

    %__MODULE__{}
    |> put_command(statement)
    |> put_format(statement)
    |> put_tokens(statement)
  end

  ################################
  # Private API
  ################################

  defp strip_newlines(raw) do
    raw
    |> String.replace("\n", " ")
    |> String.trim_leading()
    |> String.trim_trailing()
  end

  defp put_command(statement, raw) do
    command = parse_command(raw)
    %{statement | command: command}
  end

  defp parse_command("SELECT" <> _), do: :select
  defp parse_command("select" <> _), do: :select
  defp parse_command("INSERT" <> _), do: :insert
  defp parse_command("insert" <> _), do: :insert
  defp parse_command("CREATE" <> _), do: :create
  defp parse_command("create" <> _), do: :create
  defp parse_command("ALTER" <> _), do: :alter
  defp parse_command("alter" <> _), do: :alter
  defp parse_command(_), do: :unknown

  defp put_format(%{command: :insert} = statement, raw) do
    if Regex.match?(@values_regex, raw) do
      %{statement | format: "Values"}
    else
      %{statement | format: parse_format(raw)}
    end
  end

  defp put_format(statement, raw) do
    %{statement | format: parse_format(raw)}
  end

  defp parse_format(raw) do
    case Regex.run(@format_regex, raw, capture: :first) do
      [format] -> format
      _ -> nil
    end
  end

  defp put_tokens(%{command: :insert} = statement, raw) do
    if String.last(raw) == "?" do
      raw = String.trim_trailing(raw, "?")
      %{statement | tokens: [raw, "?"]}
    else
      %{statement | tokens: raw}
    end
  end

  defp put_tokens(statement, raw) do
    tokens = String.split(raw, "")
    %{statement | tokens: tokens}
  end

  defimpl String.Chars do
    def to_string(%{tokens: tokens}) do
      IO.iodata_to_binary(tokens)
    end
  end

  defimpl Inspect do
    def inspect(%{tokens: tokens}, _) do
      IO.iodata_to_binary(tokens)
    end
  end
end
